from enum import Enum

PROMPT = b">"


class Mode(Enum):
    COOL = "Cool"
    HEAT = "Heat"
    FAN = "Fan"
    DRY = "Dry"
    AUTO = "Auto"


CM_MODE_MAP = {mode.value: mode for mode in Mode}


class FanMode(Enum):
    VERY_LOW = "VLow"
    LOW = "Low"
    MEDIUM = "Med"
    HIGH = "High"
    VERY_HIGH = "Top"
    AUTO = "Auto"


# returned by "ls" line
CM_FAN_MODE_MAP = {fan_mode.value: fan_mode for fan_mode in FanMode}

# FanMode to "fspeed" argument map
CM_FAN_MODE_FSPEED_ARG_MAP = {fan_mode: fan_mode.value[0].lower() for fan_mode in FanMode}

# "properties" fan mode identifiers to FanMode
CM_PROPS_FAN_MODE_MAP = {fan_mode.value[0].lower(): fan_mode for fan_mode in FanMode}

CM_PROPS_MODE_MAP = {mode.value[0].lower(): mode for mode in Mode}


class LineType(Enum):
    DAIKIN = "DK"
    MITSUBISHI_ELECTRIC = "ME"
    MITSUBISHI_ELECTRIC_M1M2 = "M1M2"
    SANYO = "SA"
    TOSHIBA = "TO"
    PANASONIC = "PN"
    HITACHI = "HT"
    LG = "LG"
    MITSUBISHI_HEAVY_INDUSTRIES = "MH"
    GREE = "GR"
    MIDEA = "MD"
    KENTATSU = "KT"
    TRANE = "TR"
    CHIGO = "CG"
    FUJITSU = "FJ"
    SAMSUNG = "SM"
    TADIRAN_INVERTER = "TI"
    MEITAV = "MT"
    HAIER = "HA"
    BLUESTAR = "BSM"

    KNX = "KNX"

    PLUGBUS = "CH"
    PLUGBUS_WIRELESS = "CHWi"
    PLUGBUS_WIRELESS_CLIENT = "PBWi"  # from CoolPlug's side

    HDL = "HDL"

    MODBUS = "CLMB"

    MODBUS_COOLGATE4 = "CG4"
    MODBUS_COOLGATE5 = "CG5"

    MAC = "MAC"  # seen on CoolPlug

    UNUSED = "Unused"


LINE_TYPE_MAP = {item.value: item for item in LineType}


class QueryDatapoint(Enum):
    ON_OFF = "o"
    MODE = "m"
    FAN_MODE = "f"
    TARGET_TEMPERATURE = "t"
    FAILURE_CODE = "e"
    AMBIENT_TEMPERATURE = "a"
    TARGET_TEMPERATURE_PRECISE = "h"
    LOUVER_POSITION = "s"


class ModeState(Enum):
    COOL = "Cool"
    HEAT = "Heat"
    FAN = "Fan"
    DRY = "Dry"
    AUTO = "Auto"
    AUX_HEAT = "Haux"


class QueryPowerStatus(Enum):
    OFF = 0
    ON = 1


QUERY_MODE_MAPPING = {
    0: ModeState.COOL,
    1: ModeState.HEAT,
    2: ModeState.AUTO,
    3: ModeState.DRY,
    4: ModeState.AUX_HEAT,
    5: ModeState.FAN,
}


class LouverPosition(Enum):
    SWING = "a"
    STOP_SWING = "x"

    HORIZONTAL = "h"
    VERTICAL = "v"

    THIRTY_DEGREES = "3"
    FORTY_FIVE_DEGREES = "4"
    SIXTY_DEGREES = "6"


class LouverPositionState(Enum):
    SWING = "a"
    STOP_SWING = "x"

    HORIZONTAL = "h"
    VERTICAL = "v"

    THIRTY_DEGREES = "3"
    FORTY_FIVE_DEGREES = "4"
    SIXTY_DEGREES = "6"

    NOT_SUPPORTED = "0"


CM_LOUVER_POSITION_STATE_MAP = {pos.value: pos for pos in LouverPositionState}



class CoolMasterNetException(Exception):
    """
    Base class for CoolMasterNet exceptions.
    """


class RequirementNotSatisfied(CoolMasterNetException):
    """
    An extra requirement (library, etc) needs to be installed.
    """


class DeviceDisappearedException(CoolMasterNetException):
    """
    The device we were looking for is no longer on the bus.
    """


class CoolMasterNetRemoteError(CoolMasterNetException):
    """
    Base class for remote device exceptions.
    """


class UIDError:
    """
    Mixin for UID-related errors.
    """


class CoolMasterNetNoUidError(UIDError, CoolMasterNetRemoteError):
    """
    UID not found
    """

    code = 1
    verbose_code = "No UID"


class CoolMasterNetNotStrictUidError(UIDError, CoolMasterNetRemoteError):
    """
    UID must be precise
    """

    code = 2
    verbose_code = "Not Strict UID"


class CoolMasterNetBadFormatError(CoolMasterNetRemoteError):
    """
    Command format is wrong
    """

    code = 3
    verbose_code = "Bad Format"


class CoolMasterNetFailedError(CoolMasterNetRemoteError):
    """
    Command execution failed
    """

    code = 4
    verbose_code = "Failed"


class LineError:
    """
    Mixin for line-related errors.
    """


class CoolMasterNetLineUnusedError(LineError, CoolMasterNetRemoteError):
    """
    Line is unused
    """

    code = 5
    verbose_code = "Line Unused"


class CoolMasterNetUnknownCommandError(CoolMasterNetRemoteError):
    """
    Command is unknown
    """

    code = 6
    verbose_code = "Unknown Command"


class CoolMasterNetBadHvacLineError(LineError, CoolMasterNetRemoteError):
    """
    Line number is wrong
    """

    code = 7
    verbose_code = "Bad HVAC Line"


class CoolMasterNetBadFunctionError(CoolMasterNetRemoteError):
    """
    Wrong function
    """

    code = 8
    verbose_code = "Bad Function"


class CoolMasterNetBadLineTypeError(LineError, CoolMasterNetRemoteError):
    """
    Wrong line type definition
    """

    code = 9
    verbose_code = "Bad Line Type"


class CoolMasterNetBadParameterError(CoolMasterNetRemoteError):
    """
    Command parameter is wrong
    """

    code = 10
    verbose_code = "Bad Parameter"


class CoolMasterNetOkBootRequiredError(CoolMasterNetRemoteError):
    """
    Command execution will be effective after reboot
    """

    code = 11
    verbose_code = "OK, Boot Required!"


class CoolMasterNetBadGpioError(CoolMasterNetRemoteError):
    """
    Wrong GPIO
    """

    code = 12
    verbose_code = "Bad GPIO"


class CoolMasterNetSddpDisabledError(CoolMasterNetRemoteError):
    """
    SDDP module is disabled, enable it to proceed
    """

    code = 13
    verbose_code = "SDDP Disabled"


class CoolMasterNetVirtualAddressInUseError(CoolMasterNetRemoteError):
    """
    Virtual address already in use
    """

    code = 14
    verbose_code = "Virtual Address In Use"


class CoolMasterNetBadPropertyError(CoolMasterNetRemoteError):
    """
    Wrong property
    """

    code = 15
    verbose_code = "Bad Property"


class CoolMasterNetNumberOfLinesExceededError(LineError, CoolMasterNetRemoteError):
    """
    Can't define more line types
    """

    code = 16
    verbose_code = "Number of lines exceeded"


class CoolMasterNetWarningDipSwitchStateIncorrectError(CoolMasterNetRemoteError):
    """
    Dip switch state is incorrect for defined line type
    """

    code = 17
    verbose_code = "Warning! Dip Switch State Incorrect"


class CoolMasterNetSddpNotInitializedError(CoolMasterNetRemoteError):
    """
    SDDP is enabled, but Ethernet link is still down
    """

    code = 18
    verbose_code = "SDDP Not Initialized"


class ModbusError:
    """
    Mixin for Modbus-related errors.
    """


class CoolMasterNetModbusError80(ModbusError, CoolMasterNetRemoteError):
    """
    No response from the addressee
    """

    code = 80
    verbose_code = "ModBus Error:80"


class CoolMasterNetModbusError81(ModbusError, CoolMasterNetRemoteError):
    """
    Big timeout between bytes in received message
    """

    code = 81
    verbose_code = "ModBus Error:81"


class CoolMasterNetModbusError82(ModbusError, CoolMasterNetRemoteError):
    """
    Small timeout between bytes in received message
    """

    code = 82
    verbose_code = "ModBus Error:82"


class CoolMasterNetModbusError83(ModbusError, CoolMasterNetRemoteError):
    """
    Received message with internal timeout
    """

    code = 83
    verbose_code = "ModBus Error:83"


class CoolMasterNetModbusError84(ModbusError, CoolMasterNetRemoteError):
    """
    Received message is too big
    """

    code = 84
    verbose_code = "ModBus Error:84"


class CoolMasterNetModbusError85(ModbusError, CoolMasterNetRemoteError):
    """
    CRC error in received message
    """

    code = 85
    verbose_code = "ModBus Error:85"


class CoolMasterNetModbusError86(ModbusError, CoolMasterNetRemoteError):
    """
    ModBus exception in response
    """

    code = 86
    verbose_code = "ModBus Error:86"


class CoolMasterNetCollisionError(LineError, CoolMasterNetRemoteError):
    """
    Collision in sent command on HVAC line
    """

    code = 100
    verbose_code = "Collision"


class CoolMasterNetUnsupportedFeatureError(CoolMasterNetRemoteError):
    """
    Unsupported command or command's parameter for this HVAC line
    """

    code = 101
    verbose_code = "Unsupported Feature"


class CoolMasterNetIncorrectIndoorTypeError(CoolMasterNetRemoteError):
    """
    Chosen Indoor Unit doesn't support this command or command's parameter
    """

    code = 102
    verbose_code = "Incorrect Indoor Type"


class CoolMasterNetNoAckFromIndoorError(LineError, CoolMasterNetRemoteError):
    """
    Indoor Unit didn't acknowledged sent command
    """

    code = 103
    verbose_code = "No ACK From Indoor"


class CoolMasterNetTimeOutOnReceiveError(LineError, CoolMasterNetRemoteError):
    """
    No response from Indoor Unit
    """

    code = 104
    verbose_code = "Time Out on Receive"


class CoolMasterNetCsErrorInReceivedMessageError(CoolMasterNetRemoteError):
    """
    Check Sum error in received message
    """

    code = 105
    verbose_code = "CS Error In Received Message"


class CoolMasterNetLineInitInProgressError(LineError, CoolMasterNetRemoteError):
    """
    Can't show lines status due to initialization process
    """

    code = 106
    verbose_code = "Line Init In Progress..."


class CoolMasterNetLineError(LineError, CoolMasterNetRemoteError):
    """
    Some error on the HVAC line
    """

    code = 107
    verbose_code = "Line Error"


class CoolMasterNetFeedDisabledError(CoolMasterNetRemoteError):
    """
    Indoor Unit can't receive measured ambient temperature due to its switch position
    """

    code = 108
    verbose_code = "Feed Disabled"


class HDLError:
    """
    Mixin for HDL-related errors.
    """


class CoolMasterNetHdlNotInitializedError(HDLError, CoolMasterNetRemoteError):
    """
    HDL line was not defined or ethernet cable is unplugged
    """

    code = 150
    verbose_code = "HDL Not Initialized"


class CoolMasterNetHdlDbOverflowError(HDLError, CoolMasterNetRemoteError):
    """
    HDL Data Base is full, can't add new configuration
    """

    code = 151
    verbose_code = "HDL DB Overflow"


class CoolMasterNetHdlEthDisabledError(HDLError, CoolMasterNetRemoteError):
    """
    HDL over ethernet disabled
    """

    code = 152
    verbose_code = "HDL Eth Disabled"


class CoolMasterNetUidNotFoundError(UIDError, CoolMasterNetRemoteError):
    """
    Specified Indoor Unit not found in Data Base
    """

    code = 200
    verbose_code = "UID Not Found"


class CoolMasterNetStrictUidNotFoundError(UIDError, CoolMasterNetRemoteError):
    """
    Specified Indoor Unit by uid strict not found in Data Base
    """

    code = 201
    verbose_code = "Strict UID Not Found"


class CoolMasterNetIndoorRemovedError(CoolMasterNetRemoteError):
    """
    Indoor Unit removed from Data Base
    """

    code = 202
    verbose_code = "Indoor Removed"


class CoolMasterNetDbOverflowError(CoolMasterNetRemoteError):
    """
    Indoor Units Data Base is full, can't add new one
    """

    code = 203
    verbose_code = "DB Overflow"


class CoolMasterNetGroupDbOverflowError(CoolMasterNetRemoteError):
    """
    Group Data Base is full, can't add new group
    """

    code = 204
    verbose_code = "Group DB Overflow"


class CoolMasterNetVaDbOverflowError(CoolMasterNetRemoteError):
    """
    Virtual address Data Base is full, can't associate Indoor Unit with new virtual address
    """

    code = 205
    verbose_code = "VA DB Overflow"


class CoolMasterNetFdb5OverflowError(CoolMasterNetRemoteError):
    """
    Properties Data Base is full, can't add new property
    """

    code = 206
    verbose_code = "FDB5 Overflow"


class CoolMasterNetLinkDbOverflowError(CoolMasterNetRemoteError):
    """
    Link Data Base is full, can't link new CoolPlug device with Indoor Unit
    """

    code = 250
    verbose_code = "Link DB Overflow"


class CoolMasterNetNoCoolhubLineError(CoolMasterNetRemoteError):
    """
    CoolHub line not defined, define it to proceed
    """

    code = 251
    verbose_code = "No CoolHub Line"


class CoolMasterNetAutoVisibilityFailedError(CoolMasterNetRemoteError):
    """
    During link creation there was an error on adding visibility props
    """

    code = 252
    verbose_code = "Auto Visibility Failed"


class CoolMasterNetLinkAlreadyExistsError(CoolMasterNetRemoteError):
    """
    CoolPlug device already linked, delete previous link before creating the new one
    """

    code = 253
    verbose_code = "Link already exists"


class CoolMasterNetKnxDbOverflowError(CoolMasterNetRemoteError):
    """
    KNX Data Base is full, can't add new group
    """

    code = 307
    verbose_code = "KNX DB Overflow"


class KNXErrorMixin:
    """
    Mixin for KNX-related errors.
    """


class CoolMasterNetKnxNotConnectedError(KNXErrorMixin, CoolMasterNetRemoteError):
    """
    No connection with KNX chip
    """

    code = 309
    verbose_code = "KNX Not Connected"


class CoolMasterNetKnxLineNotStartedError(KNXErrorMixin, CoolMasterNetRemoteError):
    """
    KNX line not defined
    """

    code = 310
    verbose_code = "KNX Line Not Started"


class CoolMasterNetUnknownError(CoolMasterNetRemoteError):
    """
    Catch-all exception for any error we haven't got a mapping for.

    If you encounter this please open an issue so we can create a proper exception class for it.
    """


RESULT_VERBOSE_CODE_TO_EXCEPTION_MAP = {
    c.verbose_code: c for c in CoolMasterNetRemoteError.__subclasses__() if hasattr(c, "verbose_code")
}

RESULT_CODE_TO_EXCEPTION_MAP = {c.code: c for c in CoolMasterNetRemoteError.__subclasses__() if hasattr(c, "code")}



import asyncio
import logging
from typing import List, Optional, Union

from .constants import PROMPT
from .exceptions import (
    RESULT_CODE_TO_EXCEPTION_MAP,
    RESULT_VERBOSE_CODE_TO_EXCEPTION_MAP,
    CoolMasterNetUnknownError,
    RequirementNotSatisfied,
)
from .structures import UID

logger = logging.getLogger(__name__)


class BaseTransport:
    """
    A mode of connection to a CoolMasterNet device.

    Subclass this to implement the various protocols that CoolMasterNet supports - serial, TCP, REST, etc.
    """

    async def command(self, command: str) -> List[str]:
        raise NotImplementedError

    async def unit_command(self, uid: UID, command: str, extra_args: Optional[List[str]] = None):
        raise NotImplementedError

    async def plug_command(self, plug_uid: UID, command: str):
        raise NotImplementedError


class CharTransportBase(BaseTransport):
    """
    A character-based transport such as for serial or TCP-based connections.
    """

    @staticmethod
    def _parse_response(lines: List[str]) -> List[str]:
        last_line = lines[-1]

        if last_line.startswith("ERROR:"):
            result_code = int(last_line.split(":")[1])

            if result_code != 0:
                raise RESULT_CODE_TO_EXCEPTION_MAP.get(
                    result_code, CoolMasterNetUnknownError(f"Unknown result code {result_code}")
                )
        else:
            if last_line != "OK":
                raise RESULT_VERBOSE_CODE_TO_EXCEPTION_MAP.get(
                    last_line, CoolMasterNetUnknownError(f"Unknown result {lines[-1]}")
                )

        return lines[:-1]

    async def unit_command(self, uid: UID, command: str, extra_args: Optional[List[str]] = None):
        """
        Sends a command to a given UID.
        """

        args = [command, str(uid)]

        if extra_args:
            args += extra_args

        return await self.command(" ".join(args))

    async def plug_command(self, plug_uid: UID, command: str):
        """
        Sends a command to a given CoolPlug unit.
        """
        return self._parse_response(await self.unit_command(plug_uid, "plug", extra_args=[command]))


class NetworkTransportMixin:
    """
    A mixin to determine which transports rely on a network connection.
    """


class TCPTransport(NetworkTransportMixin, CharTransportBase):
    """
    A TCP transport for the "Aserver" protocol.

    TODO: this currently opens & closes the connection on each command - maybe we should persist it?
    """

    def __init__(self, host: str, port: Union[int, str] = 10102):
        self.host = host
        self.port = port

    async def command(self, command: str) -> List[str]:
        reader, writer = await asyncio.open_connection(self.host, self.port)

        try:
            await reader.readuntil(PROMPT)

            # the following is a workaround for a bug where the device will randomly
            # skip lines and instead replace them with `...Missing data...`
            # it is not a Python, asyncio or OS quirk - it is actually transmitted
            # on the wire and I have packet captures of it

            # TODO: ask CoolAutomation about this

            successful = False

            while not successful:
                writer.write((command + "\n").encode("utf-8"))
                response_bytes = await reader.readuntil(b"\r\n" + PROMPT)
                response = response_bytes.decode("utf-8")

                if "...Missing data..." in response:
                    logger.warning("Got missing data, retrying - command %s, response %s", command, response)

                    await asyncio.sleep(1)
                else:
                    successful = True

            return self._parse_response(response.split("\r\n")[:-1])
        finally:
            writer.close()
            await writer.wait_closed()

    def __str__(self):
        return f"TCP/IP {self.host}:{self.port}"


class RESTTransport(NetworkTransportMixin, BaseTransport):
    """
    A REST transport.

    URL should be the URL to the device, not including the internal /api prefix or anything,
    so if connecting to the device directly it should just be http://host:port/.
    The URL parameter is provided (as opposed to just a host/port combo) to accomodate potential
    reverse-proxies that might map the CoolMasterNet device to a subdirectory from the root URL.

    To hit the API the device's serial number is also needed.

    # TODO: actually implement this and test what's going on with the serial number
    # as I constantly got an error saying it's wrong despite providing the correct one.
    # Maybe it wants the CoolPlug serial numbers?
    """

    def __init__(self, url: str, serial_number: str):
        # TODO: figure out how to do HTTP requests with asyncio
        raise NotImplementedError


class SerialTransport(CharTransportBase):
    """
    A pyserial-backed serial transport backend.

    TODO: this currently opens & closes the port on each command.
    We should maybe persist the connection for some time?

    Example:

        s = SerialTransport('/dev/ttyS0', 9600)

        # serial to network gateway (not currently supported by pyserial-asyncio)
        s = SerialTransport('rfc2217://192.168.0.1:12345', 9600)

        # suppress errors for non-compliant gateways
        s = SerialTransport('rfc2217://192.168.0.1:12345?ign_set_control=True', 9600)

    """

    def __init__(self, url: str, baudrate=9600, **pyserial_kwargs):
        try:
            import serial
            import serial_asyncio
        except ImportError:
            raise RequirementNotSatisfied("Please install pyserial and pyserial-asyncio.")
        else:
            self._pyserial = serial
            self._pyserial_asyncio = serial_asyncio

        self.url = url
        self._baud_rate = baudrate
        self._pyserial_kwargs = pyserial_kwargs

    async def command(self, command: str) -> List[str]:
        reader, writer = await serial_asyncio.open_serial_connection(
            url=self.url, baudrate=self._baud_rate, **self._pyserial_kwargs
        )

        try:
            writer.write((command + "\n").encode("utf-8"))

            # read our own line back; note that even if echo is disabled,
            # the device still follows up any commands with at least a blank line
            # so this works either way
            await reader.readuntil()

            response = await reader.readuntil(b"\r\n" + PROMPT)
            decoded_response = response.decode("utf-8")

            return self._parse_response(decoded_response.split("\r\n")[:-1])
        finally:
            writer.close()
            await writer.wait_closed()

    def __str__(self):
        return f"Serial {self.url} @ {self._baud_rate}"



from decimal import Decimal
from typing import Any, Dict, List, Optional, Union

from .constants import (
    CM_FAN_MODE_FSPEED_ARG_MAP,
    CM_FAN_MODE_MAP,
    CM_LOUVER_POSITION_STATE_MAP,
    CM_MODE_MAP,
    CM_PROPS_FAN_MODE_MAP,
    CM_PROPS_MODE_MAP,
    FanMode,
    LineType,
    LouverPosition,
    LouverPositionState,
    Mode,
    QueryDatapoint,
)
from .data import LINE_TYPE_TO_BRAND_NAME, LINE_TYPE_TO_MODE_TEMP_RANGE
from .exceptions import CoolMasterNetNoUidError, DeviceDisappearedException
from .structures import UID, Line
from .transports import BaseTransport


class Gateway:
    """
    A CoolMasterNet gateway.
    """

    def __init__(self, transport: BaseTransport):
        self.transport = transport

    @classmethod
    async def from_transport(cls, transport: BaseTransport) -> "Gateway":
        c = cls(transport)
        await c.refresh()
        return c

    @property
    def serial_number(self):
        return self._settings.get("S/N")

    @property
    def version(self):
        return self._settings.get("Version")

    async def refresh(self):
        await self.refresh_settings()
        await self.refresh_props()
        await self.refresh_lines()
        await self.refresh_devices()

    async def refresh_devices(self):
        ls_results = await self.transport.command("ls2")

        self._devices = {}
        self._coolplug_lines = {}

        for line in ls_results:
            uid = UID.from_string(line.split()[0])

            is_coolplug = self.lines[uid.line_number].type in (LineType.PLUGBUS, LineType.PLUGBUS_WIRELESS)

            if is_coolplug:
                self._coolplug_lines[uid] = self._parse_lines_result(await self.transport.plug_command(uid, "line"))

            # according to CoolAutomation, the HVAC will always be on line 1 of a CoolPlug
            device_line = self._coolplug_lines[uid][1] if is_coolplug else self._lines[uid.line_number]

            self._devices[uid] = await Device.init_with_ls_line(
                self, line, hvac_line=device_line, properties=self.properties.get(uid, {})
            )

    async def refresh_props(self):
        props_results = await self.transport.command("props")

        heading_line = props_results[0]
        headings = list(s.strip() for s in heading_line.split("|") if s.strip())

        self._properties = {}

        for unit_line in props_results[2:]:
            props = {key: value for key, value in zip(headings, list(s.strip() for s in unit_line.split("|")))}

            self._properties[UID.from_string(props.pop("UID"))] = props

    async def refresh_lines(self):
        line_results = await self.transport.command("line")

        self._lines = self._parse_lines_result(line_results)

    async def refresh_settings(self):
        self._settings = await self.get_settings()

    async def get_settings(self) -> Dict[str, str]:
        set_results = await self.transport.command("set")

        return self._parse_key_values(set_results)

    async def get_ifconfig(self) -> Dict[str, str]:
        ifconfig_results = await self.transport.command("ifconfig")

        return self._parse_key_values(ifconfig_results)

    @staticmethod
    def _parse_key_values(lines: List[str]) -> Dict[str, str]:
        results = {}

        for line in lines:
            key, value = line.split(": ")

            results[key.strip()] = value.strip()

        return results

    @staticmethod
    def _parse_lines_result(line_results):
        lines = {}

        for heading, meta in zip(line_results[0::2], line_results[1::2]):
            line = Line.from_heading_meta(heading=heading, meta=meta)
            lines[line.number] = line

        return lines

    @property
    def properties(self) -> Dict[UID, dict]:
        return getattr(self, "_properties", {})

    @property
    def devices(self) -> Dict[UID, "Device"]:
        return getattr(self, "_devices", {})

    @property
    def lines(self) -> Dict[int, Any]:
        return getattr(self, "_lines", {})

    async def unit_command(self, uid: UID, command: str, extra_args: Optional[List[str]] = None):
        """
        Sends a command to a unit.
        """
        return await self.transport.unit_command(uid=uid, command=command, extra_args=extra_args)

    async def plug_unit_command(self, plug_uid: UID, uid: UID, command: str, extra_args: Optional[List[str]] = None):
        """
        Sends a command to a unit via a CoolPlug.
        """

        args = [command, str(uid)]

        if extra_args:
            args += extra_args

        return await self.transport.plug_command(plug_uid, " ".join(args))

    def __str__(self):
        return f"S/N: {self.serial_number} @ " + str(self.transport)


class Device:
    """
    An HVAC device connected to a gateway.

    Most methods take a "refresh" boolean parameter. This is True by default
    meaning that the internal state will be refreshed by querying the device after the command completes.

    If you are setting multiple values in quick succession, it might be better for performance
    to set it to False and then call refresh() manually afterwards.
    """

    def __init__(self, gateway: Gateway, uid: UID, hvac_line: Line, properties: dict):
        self.gateway = gateway
        self.uid = uid
        self.properties = properties
        self.hvac_line = hvac_line

    @classmethod
    async def init_with_ls_line(cls, gateway: Gateway, ls_line: str, **kwargs):
        """
        Shortcut method to create the class from an existing "ls2" result line.

        This avoids an unnecessary command being issued if the data is already available,
        for example as a result of the gateway doing an initial "ls2" when discovering devices.
        """

        c = cls(gateway=gateway, uid=UID.from_string(ls_line.split()[0]), **kwargs)

        c._populate_from_ls_line(ls_line)
        await c._refresh_louver_position()

        return c

    def __str__(self):
        return str(self.uid)

    async def command(self, command: str, extra_args=None, refresh: bool = True):
        """
        Sends a command to the device's UID.

        :param refresh: whether to refresh the internal state after completion.
        Set this to False if issuing multiple commands in sequence and then call refresh() manually.
        """

        result = await self.gateway.unit_command(uid=self.uid, command=command, extra_args=extra_args)

        if refresh:
            await self.refresh()

        return result

    async def query(self, datapoint: QueryDatapoint) -> str:
        """
        Query a datapoint and return its value as a string.
        """
        result = await self.command("query", extra_args=[datapoint.value], refresh=False)
        return result[0]

    async def refresh(self):
        """
        Refresh the internal state by querying the gateway for the device's status.
        """

        try:
            ls_line = await self.command("ls2", refresh=False)  # ["L5.001 OFF 22.0C 22.8C Top  Heat OK   - 0"]
        except CoolMasterNetNoUidError:
            raise DeviceDisappearedException(self.uid)

        self._populate_from_ls_line(ls_line[0])
        await self._refresh_louver_position()

    def _populate_from_ls_line(self, ls_line: str):
        """
        Populate the internal state from an existing line from a "ls2" command.
        """

        _, power_state, target_temp, current_temp, fan_mode, mode, error_code, filter_sign, demand = ls_line.split()

        self._power_state = power_state == "ON"
        self._temperature_unit = target_temp[-1:]
        self._target_temp = Decimal(target_temp[:-1])
        self._current_temp = Decimal(current_temp[:-1])
        self._fan_mode = CM_FAN_MODE_MAP[fan_mode]
        self._mode = CM_MODE_MAP[mode]
        self._error_code = error_code if error_code != "OK" else None
        self._filter_sign = filter_sign == "#"
        self._demand = demand != "0"

    async def _refresh_louver_position(self):
        """
        Refreshes the louver/swing position. We skip checking if the unit previously told us
        it does not support it as we assume a unit can not gain this capability at runtime.
        """

        # the hasattr check is needed as we want to allow this to be called on class initialization
        # before the internal state attributes have been created

        if not hasattr(self, "_louver_position") or self.louver_position != LouverPositionState.NOT_SUPPORTED:
            self._louver_position = CM_LOUVER_POSITION_STATE_MAP[await self.query(QueryDatapoint.LOUVER_POSITION)]

    @property
    def power_state(self) -> bool:
        """
        Boolean state of the power status of the device. True for "on", false otherwise.
        """
        return getattr(self, "_power_state")

    @property
    def temperature_unit(self) -> str:
        """
        The temperature unit set on the device. This *MAY* depend on the configuration
        of the CoolMasterNet gateway and any intermediary (CoolPlug) gateways. It is not
        guaranteed to reflect the user-facing configuration on the device's thermostat.

        Returns "C" or "F".
        """
        return getattr(self, "_temperature_unit")

    @property
    def target_temperature(self) -> Decimal:
        """
        The target temperature the device is set at.
        """
        return getattr(self, "_target_temp")

    @property
    def current_temperature(self) -> Decimal:
        """
        The current ambient temperature.

        Note that this may reflect the temperature set via set_current_temperature()
        and cached by the CoolMasterNet device but there is still no guarantee
        that the device is actually using this temperature.
        """
        return getattr(self, "_current_temp")

    @property
    def fan_mode(self) -> FanMode:
        """
        The current fan speed mode of the device.
        """
        return getattr(self, "_fan_mode")

    @property
    def mode(self) -> Mode:
        """
        The current mode (cooling, heating, drying, etc) of the device.
        """
        return getattr(self, "_mode")

    @property
    def louver_position(self) -> LouverPositionState:
        """
        The current louver position (swing) of the device.
        """
        return getattr(self, "_louver_position")

    @property
    def error_code(self) -> Optional[str]:
        """
        The current error code as reported by the device.

        None means there is no error.
        """
        return getattr(self, "_error_code")

    @property
    def filter_sign(self) -> bool:
        """
        Boolean flag whether the device is requesting a filter change.
        """
        return getattr(self, "_filter_sign")

    @property
    def demand(self) -> bool:
        """
        Boolean flag whether the device is currently reporting "demand"
        (most likely requesting refrigerant from the condensing unit).

        Not always reliable: some units or configurations consistently report this as no demand.
        """
        return getattr(self, "_demand")

    @property
    def target_temperature_range(self):
        """
        Allowed temperature range for the current mode.
        """
        ranges = self.temperature_ranges

        if ranges:
            return ranges.get(self.mode)

    @property
    def temperature_ranges(self):
        """
        A mapping of mode to temperature ranges, if known.
        """
        return LINE_TYPE_TO_MODE_TEMP_RANGE.get(self.hvac_line.type)

    @property
    def friendly_name(self) -> Optional[str]:
        """
        Returns the friendly name if one has been set using the "props" command.
        """
        return self.properties.get("Name")

    @property
    def supported_modes(self) -> List[Mode]:
        if "Modes" not in self.properties:
            return list(Mode)

        modes = (s.strip() for s in self.properties["Modes"].split() if s.strip())
        modes = list(CM_PROPS_MODE_MAP[m] for m in modes if m in CM_PROPS_MODE_MAP)

        # TODO: auxiliary heat is not supported yet
        return modes

    @property
    def supported_fan_speeds(self) -> List[FanMode]:
        if "Fspeeds" not in self.properties:
            return list(FanMode)

        modes = (s.strip() for s in self.properties["Fspeeds"].split() if s.strip())
        modes = list(CM_PROPS_FAN_MODE_MAP[m] for m in modes if m in CM_PROPS_FAN_MODE_MAP)

        return modes

    @property
    def brand_name(self) -> Optional[str]:
        """
        Returns the brand name of this unit if it can be inferred.
        """
        return LINE_TYPE_TO_BRAND_NAME.get(self.hvac_line.type)

    async def set_power_state(self, power_state: bool, refresh: bool = True):
        await self.command("on" if power_state else "off", refresh=refresh)

    async def set_mode(self, mode: Mode, refresh: bool = True):
        await self.command(mode.value.lower(), refresh=refresh)

    async def set_temperature(self, temperature: Union[int, float, Decimal], refresh: bool = True):
        await self.command("temp", [f"{temperature:.1f}"], refresh=refresh)

    async def set_fan_mode(self, fan_speed: FanMode, refresh: bool = True):
        await self.command("fspeed", [CM_FAN_MODE_FSPEED_ARG_MAP[fan_speed]], refresh=refresh)

    async def set_louver_position(self, position: LouverPosition, refresh: bool = True):
        await self.command("swing", [position.value], refresh=refresh)

    async def reset_filter_sign(self, refresh: bool = True):
        await self.command("filt", refresh=refresh)

    async def set_current_temperature(self, temperature: Union[int, float, Decimal], refresh: bool = True):
        """
        Provides the value as the current ambient temperature for the device.

        This is cached by the CoolMasterNet gateway and will then be reported as the current
        temperature measured by the device even if the device itself ignores it.
        """
        await self.command("feed", [f"{temperature:.1f}"], refresh=refresh)



from dataclasses import dataclass
from typing import Dict, List, Tuple

from .constants import LINE_TYPE_MAP, LineType


@dataclass
class UID:
    """
    Represents a CoolMasterNet strict UID.
    """

    line_number: int
    indoor_number_x: int
    indoor_number_yy: int

    @classmethod
    def from_string(cls, s: str):
        if not s.startswith("L"):
            raise ValueError('UIDs should always start with "L"')

        line, unit = s.split(".")

        return cls(line_number=int(line[1]), indoor_number_x=int(unit[0]), indoor_number_yy=int(unit[1:3]))

    def __str__(self):
        return f"L{self.line_number}.{self.indoor_number_x}{self.indoor_number_yy:02d}"

    def __hash__(self):
        return str(self).__hash__()


@dataclass
class Line:
    """
    Represents a CoolMasterNet HVAC data line.
    """

    number: int
    type: LineType
    properties: List[str]
    link_stats: Dict[str, Tuple[int, int]]

    @classmethod
    def from_heading_meta(cls, heading: str, meta: str):
        number, _type, props = cls.parse_line_heading(heading.strip())
        stats = cls.parse_line_meta(meta.strip())

        return cls(number=number, type=_type, properties=props, link_stats=stats)

    @staticmethod
    def parse_line_heading(heading: str) -> Tuple[int, LineType, List[str]]:
        # L3: CLMB Modbus Address:0x50(80) 9600_8N1
        line_number = int(heading[1])

        props = heading[3:].split()
        line_type = props[0].strip()

        return line_number, LINE_TYPE_MAP[line_type], props[1:]

    @staticmethod
    def parse_line_meta(line: str) -> Dict[str, Tuple[int, int]]:
        # Tx:7679/9204 Rx:0/0 TO:0/0 CS:0/0 Col:0/0 NAK:0/0
        results = {}

        for grouping in line.split():
            key, counts = grouping.split(":")
            count_now, count_total = counts.split("/")

            results[key.strip()] = (int(count_now), int(count_total))

        return results



from .constants import LineType, Mode

# TODO: are wireless devices supported on all lines or L5 only?
# TODO: is M1M2 on all lines or on L1 only? - I suspect L1 only on CoolPlug

LINES_TO_POSSIBLE_TYPES = {
    1: [
        LineType.DAIKIN,
        LineType.MITSUBISHI_ELECTRIC,
        LineType.MITSUBISHI_ELECTRIC_M1M2,
        LineType.SANYO,
        LineType.TOSHIBA,
        LineType.PANASONIC,
        LineType.HITACHI,
        LineType.HAIER,
    ],
    2: [
        LineType.DAIKIN,
        LineType.MITSUBISHI_ELECTRIC,
        LineType.SANYO,
        LineType.TOSHIBA,
        LineType.PANASONIC,
        LineType.HITACHI,
        LineType.HAIER,
    ],
    3: [
        LineType.LG,
        LineType.MITSUBISHI_HEAVY_INDUSTRIES,
        LineType.GREE,
        LineType.MIDEA,
        LineType.KENTATSU,
        LineType.TRANE,
        LineType.CHIGO,
        LineType.SAMSUNG,
        LineType.TADIRAN_INVERTER,
        LineType.MEITAV,
        LineType.BLUESTAR,
        LineType.KNX,
        LineType.PLUGBUS,
        LineType.PLUGBUS_WIRELESS,
        LineType.HDL,
        LineType.MODBUS_COOLGATE4,
        LineType.MODBUS_COOLGATE5,
    ],
    4: [
        LineType.LG,
        LineType.MITSUBISHI_HEAVY_INDUSTRIES,
        LineType.GREE,
        LineType.MIDEA,
        LineType.KENTATSU,
        LineType.TRANE,
        LineType.CHIGO,
        LineType.SAMSUNG,
        LineType.TADIRAN_INVERTER,
        LineType.MEITAV,
        LineType.BLUESTAR,
        LineType.HDL,
        LineType.MODBUS_COOLGATE4,
        LineType.MODBUS_COOLGATE5,
    ],
    5: [
        LineType.LG,
        LineType.MITSUBISHI_HEAVY_INDUSTRIES,
        LineType.GREE,
        LineType.MIDEA,
        LineType.KENTATSU,
        LineType.TRANE,
        LineType.CHIGO,
        LineType.SAMSUNG,
        LineType.TADIRAN_INVERTER,
        LineType.MEITAV,
        LineType.BLUESTAR,
        LineType.PLUGBUS,
        LineType.PLUGBUS_WIRELESS,
        LineType.HDL,
        LineType.MODBUS_COOLGATE4,
        LineType.MODBUS_COOLGATE5,
    ],
    6: [
        LineType.LG,
        LineType.MITSUBISHI_HEAVY_INDUSTRIES,
        LineType.GREE,
        LineType.MIDEA,
        LineType.KENTATSU,
        LineType.TRANE,
        LineType.CHIGO,
        LineType.SAMSUNG,
        LineType.TADIRAN_INVERTER,
        LineType.MEITAV,
        LineType.BLUESTAR,
        LineType.HDL,
        LineType.MODBUS_COOLGATE4,
        LineType.MODBUS_COOLGATE5,
    ],
    7: [
        LineType.LG,
        LineType.MITSUBISHI_HEAVY_INDUSTRIES,
        LineType.GREE,
        LineType.MIDEA,
        LineType.KENTATSU,
        LineType.TRANE,
        LineType.CHIGO,
        LineType.SAMSUNG,
        LineType.TADIRAN_INVERTER,
        LineType.MEITAV,
        LineType.BLUESTAR,
        LineType.HDL,
        LineType.MODBUS_COOLGATE4,
        LineType.MODBUS_COOLGATE5,
    ],
    8: [LineType.FUJITSU],
}

LINE_TYPE_TO_MODE_TEMP_RANGE = {
    LineType.MITSUBISHI_ELECTRIC: {
        Mode.COOL: range(19, 27),
        Mode.HEAT: range(17, 25),
        Mode.AUTO: range(17, 27),
        Mode.DRY: range(19, 27),
    },
    LineType.MITSUBISHI_ELECTRIC_M1M2: {
        Mode.COOL: range(19, 27),
        Mode.HEAT: range(17, 25),
        Mode.AUTO: range(17, 27),
        Mode.DRY: range(19, 27),
    },
}

LINE_TYPE_TO_BRAND_NAME = {
    LineType.DAIKIN: "Daikin",
    LineType.MITSUBISHI_ELECTRIC: "Mitsubishi Electric",
    LineType.MITSUBISHI_ELECTRIC_M1M2: "Mitsubishi Electric",
    LineType.SANYO: "Sanyo",
    LineType.TOSHIBA: "Toshiba",
    LineType.PANASONIC: "Panasonic",
    LineType.HITACHI: "Hitachi",
    LineType.LG: "LG",
    LineType.MITSUBISHI_HEAVY_INDUSTRIES: "Mitsubishi Heavy Industries",
    # all-caps on official website: https://gree.uk.com
    LineType.GREE: "GREE",
    LineType.MIDEA: "Midea",
    # all-caps on official website: https://kentatsu.eu/about-us/?lang=en
    LineType.KENTATSU: "KENTATSU",
    LineType.TRANE: "Trane",
    LineType.CHIGO: "Chigo",
    LineType.FUJITSU: "Fujitsu",
    LineType.SAMSUNG: "Samsung",
    LineType.TADIRAN_INVERTER: "Tadiran",
    LineType.MEITAV: "Meitav-tec",
    LineType.HAIER: "Haier",
    LineType.BLUESTAR: "Blue Star",
}
